if __name__ == "__main__":
    from fm import compare
    import sys

    print(sys.argv)
    with open(sys.argv[1], 'rb') as gul, open(sys.argv[2], 'rb') as fm1, open(sys.argv[3], 'rb') as fm2:
        print(compare.compare_streams(gul, fm1, fm2, precision=float(sys.argv[4]) if len(sys.argv) >= 4 else 1e-3))


__all__ = [
    'GenerateRtreeIndexData'
]

import os

from ..base import ComputationStep
from ...utils.data import get_json
from ...utils.path import as_path
from ...utils.exceptions import OasisException
from ...lookup.rtree import PerilAreasIndex


class GenerateRtreeIndexData(ComputationStep):
    """
    Generates and writes an Rtree file index of peril area IDs (area peril IDs)
    and area polygon bounds from a peril areas (area peril) file.
    """
    step_params = [
        {'name': 'index_output_file', 'flag': '-f', 'is_path': True, 'pre_exist': False,
         'help': 'Index file path (no file extension required)'},
        {'name': 'lookup_config_json', 'flag': '-m', 'is_path': True, 'pre_exist': True, 'required': True,
         'help': 'Lookup config JSON file path'},
        {'name': 'lookup_data_dir', 'flag': '-d', 'is_path': True, 'pre_exist': True,
         'help': 'Model lookup/keys data directory path'},
    ]

    def run(self):
        config = get_json(src_fp=self.lookup_config_json)
        config_dir = os.path.dirname(self.lookup_config_json) if self.lookup_config_json else self.lookup_data_dir

        peril_config = config.get('peril')

        if not peril_config:
            raise OasisException(
                'The lookup config must contain a peril-related subdictionary with a key named '
                '`peril` defining area-peril-related model information'
            )

        areas_fp = peril_config.get('file_path')

        if not areas_fp:
            raise OasisException(
                'The lookup peril config must define the path of a peril areas '
                '(or area peril) file with the key name `file_path`'
            )

        if areas_fp.startswith('%%KEYS_DATA_PATH%%'):
            areas_fp = areas_fp.replace('%%KEYS_DATA_PATH%%', self.lookup_data_dir)

        if not os.path.isabs(areas_fp):
            areas_fp = os.path.join(config_dir, areas_fp)
            areas_fp = as_path(areas_fp, 'areas_fp')

        src_type = str.lower(str(peril_config.get('file_type')) or '') or 'csv'

        peril_id_col = str.lower(str(peril_config.get('peril_id_col')) or '') or 'peril_id'

        coverage_config = config.get('coverage')

        if not coverage_config:
            raise OasisException(
                'The lookup config must contain a coverage-related subdictionary with a key named '
                '`coverage` defining coverage related model information'
            )

        coverage_type_col = str.lower(str(coverage_config.get('coverage_type_col')) or '') or 'coverage_type'

        peril_area_id_col = str.lower(str(peril_config.get('peril_area_id_col')) or '') or 'area_peril_id'

        area_poly_coords_cols = peril_config.get('area_poly_coords_cols')

        if not area_poly_coords_cols:
            raise OasisException(
                'The lookup peril config must define the column names of '
                'the coordinates used to define areas in the peril areas '
                '(area peril) file using the key `area_poly_coords_cols`'
            )

        non_na_cols = (
            tuple(col.lower() for col in peril_config['non_na_cols']) if peril_config.get('non_na_cols')
            else tuple(col.lower() for col in [peril_area_id_col] + area_poly_coords_cols.values())
        )

        col_dtypes = peril_config.get('col_dtypes') or {peril_area_id_col: int}
        sort_cols = peril_config.get('sort_cols') or peril_area_id_col
        area_poly_coords_seq_start_idx = peril_config.get('area_poly_coords_seq_start_idx') or 1
        area_reg_poly_radius = peril_config.get('area_reg_poly_radius') or 0.00166

        index_props = peril_config.get('rtree_index')
        index_props.pop('filename')

        PerilAreasIndex.create_from_peril_areas_file(
            src_fp=areas_fp,
            src_type=src_type,
            peril_id_col=peril_id_col,
            coverage_type_col=coverage_type_col,
            peril_area_id_col=peril_area_id_col,
            non_na_cols=non_na_cols,
            col_dtypes=col_dtypes,
            sort_cols=sort_cols,
            area_poly_coords_cols=area_poly_coords_cols,
            area_poly_coords_seq_start_idx=area_poly_coords_seq_start_idx,
            area_reg_poly_radius=area_reg_poly_radius,
            index_fp=self.index_output_file,
            index_props=index_props
        )
        self.logger.info('\nGenerated peril areas Rtree index file {}.idx'.format(self.index_output_file))
        self.logger.info('Generated peril areas Rtree data file {}.dat\n'.format(self.index_output_file))



from requests import codes as status
from requests import Session
from requests.adapters import HTTPAdapter
from requests_toolbelt import MultipartEncoder
from requests.exceptions import (
    ConnectionError,
    HTTPError,
    ReadTimeout,
)

from posixpath import join as urljoin
import time
import os
import logging

from ..utils.exceptions import OasisException


class APISession(Session):
    def __init__(self, api_url, username, password, timeout=25, retries=5, retry_delay=1, request_interval=0.02, logger=None, **kwargs):
        super(APISession, self).__init__(**kwargs)
        self.logger = logger or logging.getLogger()

        # Extended class vars
        self.tkn_access = None
        self.tkn_refresh = None
        self.url_base = urljoin(api_url, '')
        self.timeout = timeout
        self.retry_max = 0
        self.retry_delay = retry_delay
        self.request_interval = request_interval
        self.headers = {
            'authorization': '',
            'accept': 'application/json',
            'content-type': 'application/json',
        }
        self.adapters.clear()
        self.mount(self.url_base, HTTPAdapter(max_retries=self.retry_max))

        # Check connectivity & authentication
        self.health_check()
        self.retry_max = retries
        self.__get_access_token(username, password)

    def __get_access_token(self, username, password):
        try:
            url = urljoin(self.url_base, 'access_token/')
            r = self.post(url, json={"username": username, "password": password})
            r.raise_for_status()
            self.tkn_access = r.json()['access_token']
            self.tkn_refresh = r.json()['refresh_token']
            self.headers['authorization'] = 'Bearer {}'.format(self.tkn_access)
            return r
        except (TypeError, AttributeError, BytesWarning, HTTPError, ConnectionError, ReadTimeout) as e:
            err_msg = 'Authentication Error'
            raise OasisException(err_msg, e)

    def _refresh_token(self):
        try:
            self.headers['authorization'] = 'Bearer {}'.format(self.tkn_refresh)
            url = urljoin(self.url_base, 'refresh_token/')
            r = super(APISession, self).post(url, timeout=self.timeout)
            r.raise_for_status()

            self.tkn_access = r.json()['access_token']
            if 'refresh_token' in r.json():
                self.logger.debug(f"Refreshing access token")
                self.tkn_refresh = r.json()['refresh_token']
            self.headers['authorization'] = 'Bearer {}'.format(self.tkn_access)
            return r
        except (TypeError, AttributeError, BytesWarning, HTTPError, ConnectionError, ReadTimeout) as e:    
            err_msg = 'Authentication Error: {}'.format(r.text)
            raise OasisException(err_msg, e)

    def unrecoverable_error(self, error, msg=None):
        err_r = error.response
        err_msg = 'api error: {}, url: {}, msg: {}'.format(err_r.status_code, err_r.url, err_r.text)
        self.logger.error(err_msg)
        if msg:
            self.logger.error(msg)

    # Connection Error Handlers
    def __recoverable(self, error, url, request, counter=1):
        self.logger.debug(f"Connection exception handler: {error}")
        if not (counter < self.retry_max):
            self.logger.debug("Max retries of '{}' reached".format(self.retry_max))
            raise OasisException(f'Failed to recover from error: {error}', error)

        if isinstance(error, (ConnectionError, ReadTimeout)):
            self.logger.debug(f"Recoverable error [{error}] from {request} {url}")
            self.logger.debug(f"Backoff timer: {self.retry_delay * counter}")

            # Reset HTTPAdapter & clear connection pool
            self.adapters.clear()
            self.mount(self.url_base, HTTPAdapter(max_retries=self.retry_max))
            time.sleep(self.retry_delay * counter)
            return True

        elif isinstance(error, HTTPError):
            http_err_code = error.response.status_code
            self.logger.debug(f"Recoverable error [{error}] from {request} {url}")

            if http_err_code in [502, 503, 504]:
                error = "HTTP {}".format(http_err_code)
                return True
            elif http_err_code in [401]:
                self.logger.debug(f"requesting refresh token")
                self._refresh_token()
                return True
        return False

    # @oasis_log
    def health_check(self):
        """
        Checks the health of the server.

        """
        try:
            url = urljoin(self.url_base, 'healthcheck/')
            return super(APISession, self).get(url)
        except (TypeError, AttributeError, BytesWarning, HTTPError, ConnectionError, ReadTimeout) as e:
            err_msg = 'Health check failed: Unable to connect to {}'.format(self.url_base)
            raise OasisException(err_msg, e)

    def upload(self, url, filepath, content_type, **kwargs):
        counter = 0
        while True:
            counter += 1
            try:
                abs_fp = os.path.realpath(os.path.expanduser(filepath))
                m = MultipartEncoder(fields={'file': (os.path.basename(filepath), open(abs_fp, 'rb'), content_type)})
                r = super(APISession, self).post(url, data=m, headers={'Content-Type': m.content_type}, timeout=self.timeout, **kwargs)
                r.raise_for_status()
                time.sleep(self.request_interval)
            except (HTTPError, ConnectionError, ReadTimeout) as e:
                if self.__recoverable(e, url, 'GET', counter):
                    continue
                else:
                    self.logger.debug(f'Unrecoverable error: {e}')
                    raise e
            return r

    def get(self, url, **kwargs):
        counter = 0
        while True:
            counter += 1
            try:
                r = super(APISession, self).get(url, timeout=self.timeout, **kwargs)
                r.raise_for_status()
                time.sleep(self.request_interval)
            except (HTTPError, ConnectionError, ReadTimeout) as e:
                if self.__recoverable(e, url, 'GET', counter):
                    continue
                else:
                    self.logger.debug(f'Unrecoverable error: {e}')
                    raise e
            return r

    def post(self, url, **kwargs):
        counter = 0
        while True:
            counter += 1
            try:
                r = super(APISession, self).post(url, timeout=self.timeout, **kwargs)
                r.raise_for_status()
                time.sleep(self.request_interval)
            except (HTTPError, ConnectionError, ReadTimeout) as e:
                if self.__recoverable(e, url, 'POST', counter):
                    continue
                else:
                    self.logger.debug(f'Unrecoverable error: {e}')
                    raise e
            return r

    def delete(self, url, **kwargs):
        counter = 0
        while True:
            counter += 1
            try:
                r = super(APISession, self).delete(url, timeout=self.timeout, **kwargs)
                r.raise_for_status()
                time.sleep(self.request_interval)
            except (HTTPError, ConnectionError, ReadTimeout) as e:
                if self.__recoverable(e, url, 'DELETE', counter):
                    continue
                else:
                    self.logger.debug(f'Unrecoverable error: {e}')
                    raise e
            return r

    def put(self, url, **kwargs):
        counter = 0
        while True:
            counter += 1
            try:
                r = super(APISession, self).put(url, timeout=self.timeout, **kwargs)
                r.raise_for_status()
                time.sleep(self.request_interval)
            except (HTTPError, ConnectionError, ReadTimeout) as e:
                if self.__recoverable(e, url, 'OPTIONS', counter):
                    continue
                else:
                    self.logger.debug(f'Unrecoverable error: {e}')
                    raise e
            return r



__all__ = [
    "HelperTabComplete",
]


from ..base import ComputationStep

import os
import sys
from platform import system


class HelperTabComplete(ComputationStep):
    """
    Adds required command to `.bashrc` Linux or .bash_profile for mac
    so that Command autocomplete works for oasislmf CLI
    """
    step_params = [
        {'name': 'bash_rc_file', 'flag': '-p', 'help': 'Path to bash configuration RC file, "~/.bashrc". '},
        {'name': 'no_confirm', 'flag': '-y', 'action': 'store_true', 'default': False, 'help': 'Skip the confirmation prompt'},
    ]

    def confirm_action(self, question_str, no_confirm=False):
        self.logger.debug('Prompt user for confirmation')
        if no_confirm:
            return True
        try:
            check = str(input("%s (Y/N): " % question_str)).lower().strip()
            if check[:1] == 'y':
                return True
            elif check[:1] == 'n':
                return False
            else:
                self.logger.error('Enter "y" for Yes, "n" for No or Ctrl-C to exit.\n')
                return self.confirm_action(question_str)
        except KeyboardInterrupt:
            self.logger.error('\nexiting.')

    def install_autocomplete(self, target_file=None):
        msg_success = 'Auto-Complete installed.'
        msg_failed = 'install failed'
        msg_installed = 'Auto-Complete feature is already enabled.'
        msg_reload_bash = '\n To activate reload bash by running: \n     source {}'.format(target_file)
        cmd_header = '# Added by OasisLMF\n'
        cmd_autocomplete = 'complete -C completer_oasislmf oasislmf\n'

        try:
            if os.path.isfile(target_file):
                # Check command is in file
                with open(target_file, "r") as rc:
                    if cmd_autocomplete in rc.read():
                        self.logger.info(msg_installed)
                        self.logger.info(msg_reload_bash)
                        sys.exit(0)
            else:
                # create new file at set location
                basedir = os.path.dirname(target_file)
                if not os.path.isdir(basedir):
                    os.makedirs(basedir)

            # Add complete command
            with open(target_file, "a") as rc:
                rc.write(cmd_header)
                rc.write(cmd_autocomplete)
                self.logger.info(msg_success)
                self.logger.info(msg_reload_bash)
        except Exception as e:
            self.logger.error('{}: {}'.format(msg_failed, e))

    def run(self):

        # select default bashrc if not set
        if not self.bash_rc_file:
            default_file = '.bash_profile' if system == 'Darwin' else '.bashrc'
            self.bash_rc_file = os.path.join(
                os.path.expanduser('~'),
                default_file
            )

        # Prompt user, and then install
        msg_user = 'Running this will append a command to the following file:\n'
        if self.confirm_action("{} {}".format(msg_user, self.bash_rc_file), self.no_confirm):
            self.install_autocomplete(self.bash_rc_file)



from .policy import calc
from .common import float_equal_precision, np_oasis_float

from numba import njit
import numpy as np
import os
import logging
logger = logging.getLogger(__name__)


@njit(cache=True, fastmath=True, error_model="numpy")
def back_allocate(node, children, nodes_array, losses, loss_indexes, loss_i, computes, next_compute_i, child_loss_pointer):
    len_children = children[node['children']]
    if len_children > 1:
        for layer in range(node['layer_len']):
            proportion = np.empty_like(losses[0])
            use_loss = np.zeros_like(losses[0], dtype=np.uint8)
            ba_loss = losses[loss_indexes[node['ba'] + layer]]
            sum_loss = losses[loss_indexes[node['loss'] + layer]]
            for i in range(proportion.shape[0]):
                if ba_loss[i] < float_equal_precision:
                    proportion[i] = 0
                else:
                    # in case of max deductible sum_loss can be 0 while ba_loss > 0
                    # the historical hack is to then use the original loss of the child instead of the il
                    # TODO: this need to be reworked to avoid loss inconsistencies at the limit sum loss = 0
                    #       when max deductible is triggered (loss reallocated proportionally to children's under_limit)
                    if sum_loss[i] < float_equal_precision:
                        use_loss[i] = 1
                        for c in range(node['children'] + 1, node['children'] + len_children + 1):
                            sum_loss[i] += losses[loss_indexes[nodes_array[children[c]]['loss'] + layer]][i]
                        if sum_loss[i] < float_equal_precision:
                            proportion[i] = 0
                            continue
                    proportion[i] = ba_loss[i] / sum_loss[i]

            for c in range(node['children'] + 1, node['children'] + len_children + 1):
                child = nodes_array[children[c]]
                child_ba = losses[loss_i]
                child_loss = losses[loss_indexes[child['loss'] + layer]]
                child_il = losses[loss_indexes[child[child_loss_pointer] + layer]]
                for i in range(proportion.shape[0]):
                    if use_loss[i]:
                        child_ba[i] = proportion[i] * child_loss[i]
                    else:
                        child_ba[i] = proportion[i] * child_il[i]
                loss_indexes[child['ba'] + layer], loss_i = loss_i, loss_i + 1

        for c in range(node['children'] + 1, node['children'] + len_children + 1):
            computes[next_compute_i], next_compute_i = children[c], next_compute_i + 1

    else:
        computes[next_compute_i], next_compute_i = children[node['children'] + 1], next_compute_i + 1
        child = nodes_array[children[node['children'] + 1]]
        loss_indexes[child['ba']: child['ba'] + node['layer_len']] = loss_indexes[node['ba']: node['ba'] + node['layer_len']]

    return loss_i, next_compute_i


@njit(cache=True, fastmath=True)
def compute_event(compute_info,
                  net_loss,
                  nodes_array,
                  node_parents_array,
                  node_profiles_array,
                  losses,
                  loss_indexes,
                  extras,
                  extra_indexes,
                  children,
                  computes,
                  next_compute_i,
                  fm_profile,
                  stepped):

    loss_i = next_compute_i
    extra_i = 0
    compute_i = 0
    for level in range(compute_info['start_level'], compute_info['max_level'] + 1):#perform the bottom up loss computation
        # print(level, next_compute_i, compute_i, next_compute_i-compute_i, computes[compute_i:compute_i+2], computes[next_compute_i-2: next_compute_i+1])
        next_compute_i += 1 # next_compute_i will stay at 0 wich will stop the while loop and start the next level
        while computes[compute_i]:
            node, compute_i = nodes_array[computes[compute_i]], compute_i + 1

            #compute loss sums and extra sum
            len_children = children[node['children']]
            if len_children:#if no children and in compute then layer 1 loss is already set
                if len_children > 1:
                    for p in range(node['profile_len']):
                        node_loss = losses[loss_i]
                        loss_indexes[node['loss'] + p], loss_i = loss_i, loss_i + 1
                        node_extra = extras[extra_i]
                        extra_indexes[node['extra'] + p], extra_i = extra_i, extra_i + 1

                        node_loss.fill(0)
                        node_extra.fill(0)
                        for c in range(node['children'] + 1, node['children'] + len_children + 1):
                            child = nodes_array[children[c]]
                            node_loss += losses[loss_indexes[child['il'] + p]]
                            node_extra += extras[extra_indexes[child['extra'] + p]]
                    #fill up all layer if necessary
                    for layer in range(node['profile_len'], node['layer_len']):
                        loss_indexes[node['loss'] + layer] = loss_indexes[node['loss']]
                        extra_indexes[node['extra'] + layer] = extra_indexes[node['extra']]
                else:
                    child = nodes_array[children[node['children'] + 1]]
                    loss_indexes[node['loss']:node['loss'] + node['layer_len']] = loss_indexes[child['il']:child['il'] + node['layer_len']]

                    for p in range(node['profile_len']):
                        extras[extra_i][:] = extras[extra_indexes[child['extra'] + p]]
                        extra_indexes[node['extra'] + p], extra_i = extra_i, extra_i + 1

                    #fill up all layer if necessary
                    for layer in range(node['profile_len'], node['layer_len']):
                        extra_indexes[node['extra'] + layer] = extra_indexes[node['extra']]

            else:
                for p in range(node['profile_len']):
                    extras[extra_i].fill(0)
                    extra_indexes[node['extra'] + p], extra_i = extra_i, extra_i + 1

                #fill up all layer if necessary
                for layer in range(node['profile_len'], node['layer_len']):
                    extra_indexes[node['extra'] + layer] = extra_indexes[node['extra']]

            #compute il
            for p in range(node['profile_len']):
                node_profile = node_profiles_array[node['profiles']+p]
                if node_profile['i_start'] < node_profile['i_end']:
                    extra = extras[extra_indexes[node['extra'] + p]]
                    loss_in = losses[loss_indexes[node['loss'] + p]]
                    for profile_index in range(node_profile['i_start'], node_profile['i_end']):
                        calc(fm_profile[profile_index],
                             losses[loss_i],
                             loss_in,
                             extra[0],
                             extra[1],
                             extra[2],
                             stepped)
                    loss_indexes[node['il'] + p], loss_i = loss_i, loss_i + 1
                else:
                    loss_indexes[node['il'] + p] = loss_indexes[node['loss'] + p]

            for layer in range(node['profile_len'], node['layer_len']):
                loss_indexes[node['il'] + layer] = loss_indexes[node['il']]

            if level == compute_info['max_level']:
                loss_indexes[node['ba']:node['ba'] + node['layer_len']] = loss_indexes[node['il']:node['il'] + node['layer_len']]
                if node['parent_len']:# for allocation 1 we set only the parent in computes
                    parent_id = node_parents_array[node['parent']]
                    computes[next_compute_i], next_compute_i = parent_id, next_compute_i + 1
                else:
                    computes[next_compute_i], next_compute_i = computes[compute_i - 1], next_compute_i + 1
            else:
                # set parent children and next computation
                for pa in range(node['parent_len']):
                    parent_id = node_parents_array[node['parent']+pa]
                    parent = nodes_array[parent_id]
                    parent_children_len = children[parent['children']] + 1
                    if parent_children_len == 1 and not pa:#only parent 0 is a real parent, parent 2 is only used on alloc 1
                        computes[next_compute_i], next_compute_i = parent_id, next_compute_i + 1
                    children[parent['children']] = parent_children_len
                    children[parent['children'] + parent_children_len] = computes[compute_i - 1]
        compute_i += 1

    if compute_info['allocation_rule'] == 0:
        pass
    elif compute_info['allocation_rule'] == 1:
        next_compute_i += 1
        while computes[compute_i]:
            node, compute_i = nodes_array[computes[compute_i]], compute_i + 1
            len_children = children[node['children']]
            if len_children > 1:
                # we are summing the input loss of level 0 or 1 so there is no layer to take into account
                node_loss = losses[loss_i]
                loss_indexes[node['loss']], loss_i = loss_i, loss_i + 1

                node_loss.fill(0)
                for c in range(node['children'] + 1, node['children'] + len_children + 1):
                    child = nodes_array[children[c]]
                    node_loss += losses[loss_indexes[child['loss']]]
                for layer in range(1, node['layer_len']):
                    loss_indexes[node['loss'] + layer] = loss_indexes[node['loss']]
            loss_i, next_compute_i = back_allocate(node, children, nodes_array, losses, loss_indexes, loss_i, computes, next_compute_i, 'loss')
        compute_i += 1
    else:
        for level in range(compute_info['max_level'] - compute_info['start_level']):# perform back allocation 2
            # print(level, next_compute_i, compute_i, next_compute_i-compute_i, computes[compute_i:compute_i + 2], computes[next_compute_i - 1: next_compute_i + 1])
            next_compute_i += 1
            while computes[compute_i]:
                node, compute_i = nodes_array[computes[compute_i]], compute_i + 1
                loss_i, next_compute_i = back_allocate(node,
                                                       children,
                                                       nodes_array,
                                                       losses,
                                                       loss_indexes,
                                                       loss_i,
                                                       computes,
                                                       next_compute_i,
                                                       'il')
            compute_i += 1
    if net_loss:
        # we go through the last level node and replace the gross loss by net loss, then reset compute_i to its value
        out_compute_i = compute_i
        while computes[compute_i]:
            node_i, compute_i= computes[compute_i], compute_i + 1
            node = nodes_array[node_i]
            # net loss layer i is initial loss - sum of all layer up to i
            losses[loss_indexes[node['ba']]] = np.maximum(losses[loss_indexes[node['loss']]] - losses[loss_indexes[node['ba']]], 0)
            for layer in range(1, node['layer_len']):
                losses[loss_indexes[node['ba'] + layer]] = np.maximum((losses[loss_indexes[node['ba'] + layer - 1 ]]
                                                                       - losses[loss_indexes[node['ba'] + layer]]),
                                                                      0)

        compute_i = out_compute_i

    # print(compute_info['max_level'], next_compute_i, compute_i, next_compute_i-compute_i, computes[compute_i:compute_i + 2], computes[next_compute_i - 1: next_compute_i + 1])
    return compute_i, loss_i, extra_i


def init_variable(compute_info, len_sample, temp_dir, low_memory):
    if low_memory:
        losses = np.memmap(os.path.join(temp_dir, "temp_loss.bin"), mode='w+',
                           shape=(compute_info['loss_len'], len_sample), dtype=np_oasis_float)
        extras = np.memmap(os.path.join(temp_dir, "extras.bin"), mode='w+',
                           shape=(compute_info['extra_len'], 3, len_sample), dtype=np_oasis_float)
    else:
        losses = np.zeros((compute_info['loss_len'], len_sample), dtype=np_oasis_float)
        extras = np.zeros((compute_info['loss_len'], 3, len_sample), dtype=np_oasis_float)

    loss_indexes = np.zeros(compute_info['loss_len'], dtype=np.uint32)
    extra_indexes = np.zeros(compute_info['extra_len'], dtype=np.uint32)
    children = np.zeros(compute_info['children_len'], dtype=np.uint32)
    computes = np.zeros(compute_info['compute_len'], dtype=np.uint32)

    return losses, loss_indexes, extras, extra_indexes, children, computes


@njit(cache=True)
def reset_variable(children, compute_i, computes, loss_i, losses):
    computes[:compute_i].fill(0)
    losses[:loss_i].fill(0)
    children.fill(0)



#!/usr/bin/env python
import argparse

import select
import numpy as np
from numba import njit
import time
from io import BytesIO
import queue
import concurrent.futures

import logging
logger = logging.getLogger(__name__)


last_event_padding = b'\x00\x00\x00\x00\x00\x00\x00\x00'
number_size = 4
CHECK_STOPPER_DURATION = 1


parser = argparse.ArgumentParser()
parser.add_argument('-i', '--pipe-in', help='names of the input file_path', nargs='+')
parser.add_argument('-o', '--pipe-out', help='names of the output file_path', nargs='+')
parser.add_argument('-r', '--read-size', help='maximum size of chunk read from input', default=1_048_576, type=int)
parser.add_argument('-w', '--write-size', help='maximum size of chunk read from input', default=1024, type=int)
parser.add_argument('-q', '--queue-size', help='maximum size of the queue', default=50, type=int)
parser.add_argument('-v', '--logging-level', help='logging level (debug:10, info:20, warning:30, error:40, critical:50',
                    default=30, type=int)


class ProducerStopped(RuntimeError):
    pass


@njit(cache=True)
def get_next_event_index(read_buffer, last_event_index, last_event_id, max_cursor):
    """
    try to get the index of the end of the event
    if found return the index and 0 to indicate it is found
    if not found return the index of the last item parsed and the last event id

    :param sub: byte array to parse
    :param last_item_index: last index parsed
    :param last_event_id: last event idea parsed (0 means no event)
    :return: last index parsed, last event idea parsed (0 means the chunk sub[:last_item_index] is a full event
    """
    cursor = last_event_index
    while cursor < max_cursor - 4:
        cur_event_id = read_buffer[cursor]
        if last_event_id != cur_event_id:
            if last_event_id == 0:
                last_event_id = read_buffer[cursor]
            else:
                return cursor, last_event_id, 1

        cursor += 2
        while cursor < max_cursor - 2:
            sidx = read_buffer[cursor]
            cursor += 2
            if sidx == 0:
                last_event_index = cursor
                break
    return last_event_index, last_event_id, 0


def produce(even_queue, event, stopper):
    while not stopper[0]:
        try:
            even_queue.put(event, timeout=CHECK_STOPPER_DURATION)
            break
        except queue.Full:
            pass
    else:
        raise ProducerStopped()


def producer(in_stream, pipeline, read_size, stopper):
    read_buffer = memoryview(bytearray(read_size))
    buf_as_int32 = np.ndarray(read_size // number_size, buffer=read_buffer, dtype=np.int32)
    event_buffer = BytesIO()
    left_over = 0
    last_event_index = 0
    last_event_id = 0

    wait_read_time = 0
    read_input_time = 0
    parse_input_time = 0
    buffer_management_time = 0

    while True:
        tw = time.time()
        select.select([in_stream], [], [])
        tr = time.time()
        wait_read_time += tr - tw
        len_read = in_stream.readinto1(read_buffer[left_over:])
        read_input_time += time.time() - tr
        if not len_read:
            in_stream.close()
            event_buffer.write(read_buffer[:left_over])
            if read_buffer[left_over - 8:left_over-4] != b'\x00\x00\x00\x00':
                event_buffer.write(last_event_padding)
            event_buffer.seek(0)
            produce(pipeline, event_buffer, stopper)
            break

        valid_buf = len_read + left_over

        while True:
            tp = time.time()
            event_index, last_event_id, event_finished = get_next_event_index(buf_as_int32, last_event_index, last_event_id,
                                                                              valid_buf // number_size)
            tm = time.time()
            parse_input_time += tm - tp

            event_buffer.write(read_buffer[last_event_index * number_size: event_index * number_size])

            if event_finished:
                event_buffer.seek(0)
                produce(pipeline, event_buffer, stopper)
                event_buffer = BytesIO()
                last_event_index = event_index
                last_event_id = 0
            else:
                left_over = valid_buf - number_size * event_index
                read_buffer[:left_over] = read_buffer[number_size * event_index: valid_buf]
                last_event_index = 0
                break
            buffer_management_time += time.time() - tm

    return wait_read_time, read_input_time, parse_input_time, buffer_management_time


def consumer(out_stream, pipeline, write_size, sentinel, stopper):
    s_tot=0
    w_tot=0
    p_tot=0
    while True:
        tp = time.time()
        event_buf = pipeline.get()
        p_tot += time.time() - tp
        if event_buf is sentinel:
            break
        else:
            while True:
                data = event_buf.read(write_size)
                if not data:
                    break
                ts = time.time()
                select.select([], [out_stream], [])
                tw = time.time()
                s_tot += tw-ts
                try:
                    out_stream.write(data)
                except:
                    stopper[0] = True
                    raise
                w_tot += time.time() - tw
    return s_tot, w_tot, p_tot


def balance(pipe_in, pipe_out, read_size, write_size, queue_size):
    """
    Load balance events for a list of input fil_path to a list of output fil_path

    :param pipe_in: list of fil_path
        fil_path to take as input
    :param pipe_out: list of fil_path
        fil_path to take as input
    :param read_size: int
        size of the maximum amount of Byte read from one input at a time
    :param queue_size: int
        maximum size ofthe buffer queue

    """
    inputs = [open(p, 'rb') for p in pipe_in]
    outputs = [open(p, 'wb') for p in pipe_out]

    pipeline = queue.Queue(maxsize=queue_size)
    sentinel = object()
    stopper = np.zeros(1, dtype=np.bool)
    try:
        # check stream input header and write it to the stream output
        headers = set([s.read(8) for s in inputs])
        if len(headers) != 1:
            raise Exception('input streams have different header type')
        header = headers.pop()
        [s.write(header) for s in outputs]
        with concurrent.futures.ThreadPoolExecutor(max_workers=len(inputs)+len(outputs)) as executor:
            producer_task = [executor.submit(producer, s, pipeline, read_size, stopper) for s in inputs]
            consumer_task = [executor.submit(consumer, s, pipeline, write_size, sentinel, stopper) for s in outputs]
            try:
                prod_t = [t.result() for t in producer_task]
            finally:
                for _ in pipe_out:
                    pipeline.put(sentinel)

            wait_read_time, read_input_time, parse_input_time, buffer_management_time = 0, 0, 0, 0
            for t in prod_t:
                wait_read_time += t[0]
                read_input_time += t[1]
                parse_input_time += t[2]
                buffer_management_time += t[3]

            cons_t = [t.result() for t in consumer_task]

            wait_write_time, write_output_time, wait_pipeline = 0, 0, 0
            for t in cons_t:
                wait_write_time += t[0]
                write_output_time += t[1]
                wait_pipeline += t[2]

            logger.info(f"""
    wait_read_time = {wait_read_time}, {wait_read_time / len(inputs)}
    read_input_time = {read_input_time}, {read_input_time / len(inputs)}
    parse_input_time = {parse_input_time}, {parse_input_time / len(inputs)}
    buffer_management_time = {buffer_management_time}, {buffer_management_time / len(inputs)}
    wait_write_time = {wait_write_time}, {wait_write_time / len(outputs)}
    write_output_time = {write_output_time}, {write_output_time / len(outputs)}
    wait_pipeline = {wait_pipeline}, {wait_pipeline / len(outputs)}
    """)


    finally:
        [s.close() for s in inputs]
        [s.close() for s in outputs]


def main():
    kwargs = vars(parser.parse_args())

    # add handler to fm logger
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logging_level = kwargs.pop('logging_level')
    logger.setLevel(logging_level)

    balance(**kwargs)


if __name__ == '__main__':
    main()



from .policy import calc
from .common import np_oasis_float, np_oasis_int, EXTRA_VALUES

from numba import njit
import numpy as np
import os
import logging
logger = logging.getLogger(__name__)


@njit(cache=True)
def merge_2_sorted_array(storage_array_a, ptr_a, end_a, storage_array_b, ptr_b, end_b, storage_array_res, ptr_res):
    while ptr_a < end_a and ptr_b < end_b:
        if storage_array_a[ptr_a] == storage_array_b[ptr_b]:
            storage_array_res[ptr_res] = storage_array_a[ptr_a]
            ptr_a += 1
            ptr_b += 1

        elif storage_array_a[ptr_a] < storage_array_b[ptr_b]:
            storage_array_res[ptr_res] = storage_array_a[ptr_a]
            ptr_a += 1

        else:
            storage_array_res[ptr_res] = storage_array_b[ptr_b]
            ptr_b += 1
        ptr_res += 1

    while ptr_a < end_a:
        storage_array_res[ptr_res] = storage_array_a[ptr_a]
        ptr_a += 1
        ptr_res += 1

    while ptr_b < end_b:
        storage_array_res[ptr_res] = storage_array_b[ptr_b]
        ptr_b += 1
        ptr_res += 1
    return ptr_res


@njit(cache=True)
def single_array_to_sparse(sidx, sidx_ptr_i, sidx_val, temp_node_sidx,
                           loss_ptr_i, loss_val, temp_node_loss,
                           extras_ptr_i, extras_val, temp_node_extras):
    """
    sidx: values to put in the sparse array
    sidx_ptr_i: index of the next sparse sidx value
    sidx_val: sparse array to store the sidx
    temp_node_sidx: dense array of sidx
        """
    sidx_val[sidx_ptr_i] = sidx
    temp_node_sidx[sidx] = False

    loss_val[loss_ptr_i] = temp_node_loss[sidx]
    temp_node_loss[sidx] = 0

    extras_val[extras_ptr_i] = temp_node_extras[sidx]
    temp_node_extras[sidx] = 0
    return sidx_ptr_i + 1, loss_ptr_i + 1, extras_ptr_i + 1,


@njit(cache=True, fastmath=True, error_model="numpy")
def back_allocate(node, children, nodes_array, sidx_indexes, sidx_indptr, sidx_val,
                  loss_indptr, loss_val, loss_ptr_i,
                  computes, next_compute_i, child_loss_pointer, temp_proportion):
    len_children = children[node['children']]
    if len_children > 1:
        node_sidx_start = sidx_indptr[sidx_indexes[node['node_id']]]
        node_sidx_end = sidx_indptr[sidx_indexes[node['node_id']] + 1]

        for layer in range(node['layer_len']):
            node_ba_loss_cur = loss_indptr[node['ba'] + layer]
            node_sum_loss_cur = loss_indptr[node['loss'] + layer]
            # print(node['node_id'], 'node_ba_loss_cur', (node_sidx_end - node_sidx_start), loss_val[node_ba_loss_cur:node_ba_loss_cur + node_sidx_end - node_sidx_start])
            for node_sidx_cur in range(node_sidx_start, node_sidx_end):
                sidx = sidx_val[node_sidx_cur]
                if loss_val[node_ba_loss_cur]:
                    if loss_val[node_sum_loss_cur]:
                        temp_proportion[sidx] = loss_val[node_ba_loss_cur] / loss_val[node_sum_loss_cur]
                    else:
                        sum_loss = 0
                        for c in range(node['children'] + 1, node['children'] + len_children + 1):
                            child = nodes_array[children[c]]
                            child_sidx_i = np.searchsorted(sidx_val[sidx_indptr[sidx_indexes[child['node_id']]]:sidx_indptr[sidx_indexes[child['node_id']] + 1]], sidx)
                            if sidx == sidx_val[sidx_indptr[sidx_indexes[child['node_id']]] + child_sidx_i]:
                                sum_loss += loss_val[loss_indptr[child['loss'] + layer] + child_sidx_i]
                        if sum_loss:
                            temp_proportion[sidx] = - loss_val[node_ba_loss_cur] / sum_loss
                        else:
                            temp_proportion[sidx] = 0
                else:
                    temp_proportion[sidx] = 0
                node_ba_loss_cur += 1
                node_sum_loss_cur += 1

            for c in range(node['children'] + 1, node['children'] + len_children + 1):
                child = nodes_array[children[c]]
                child_sidx_start = sidx_indptr[sidx_indexes[child['node_id']]]
                child_sidx_end = sidx_indptr[sidx_indexes[child['node_id']] + 1]
                child_loss_cur = loss_indptr[child['loss'] + layer]
                child_il_cur = loss_indptr[child[child_loss_pointer] + layer]
                loss_indptr[child['ba'] + layer] = loss_ptr_i
                for child_sidx_cur in range(child_sidx_start, child_sidx_end):
                    proportion = temp_proportion[sidx_val[child_sidx_cur]]
                    if proportion > 0:
                        loss_val[loss_ptr_i] = loss_val[child_il_cur] * proportion
                    elif proportion == 0:
                        loss_val[loss_ptr_i] = 0
                    else:
                        loss_val[loss_ptr_i] = - loss_val[child_loss_cur] * proportion
                    child_loss_cur +=1
                    child_il_cur += 1
                    loss_ptr_i += 1

        for c in range(node['children'] + 1, node['children'] + len_children + 1):
            computes[next_compute_i], next_compute_i = children[c], next_compute_i + 1

    else:
        computes[next_compute_i], next_compute_i = children[node['children'] + 1], next_compute_i + 1
        child = nodes_array[children[node['children'] + 1]]
        loss_indptr[child['ba']: child['ba'] + node['layer_len']] = loss_indptr[node['ba']: node['ba'] + node['layer_len']]

    return loss_ptr_i, next_compute_i


@njit(cache=True, fastmath=True)
def compute_event(compute_info,
                  net_loss,
                  nodes_array,
                  node_parents_array,
                  node_profiles_array,
                  len_array, max_sidx_val, sidx_indexes, sidx_indptr, sidx_val, loss_indptr, loss_val, extras_indptr, extras_val,
                  children,
                  computes,
                  next_compute_i,
                  fm_profile,
                  stepped):
    # len_array = computation_structure.len_array
    # max_sidx_val = computation_structure.max_sidx_val
    # sidx_indexes = computation_structure.sidx_indexes
    # sidx_indptr = computation_structure.sidx_indptr
    # sidx_val = computation_structure.sidx_val
    # loss_indptr = computation_structure.loss_indptr
    # loss_val = computation_structure.loss_val
    # extras_indptr = computation_structure.extras_indptr
    # extras_val = computation_structure.extras_val
    # at this point we have as many loss as sidx as there is only 1 layer and 1 loss type 'loss'
    sidx_ptr_i = sidx_indptr[next_compute_i]
    loss_ptr_i = sidx_ptr_i
    extras_ptr_i = 0
    sidx_i = next_compute_i
    compute_i = 0

    temp_node_sidx = np.zeros(len_array, dtype=np.uint8)
    temp_node_loss = np.zeros(len_array, dtype=np_oasis_float)
    temp_node_extras = np.zeros((len_array, 3), dtype=np_oasis_float)

    for level in range(compute_info['start_level'], compute_info['max_level'] + 1):#perform the bottom up loss computation
        # print(level, next_compute_i, compute_i, next_compute_i-compute_i, computes[compute_i:compute_i+2], computes[next_compute_i-2: next_compute_i+1])
        next_compute_i += 1 # next_compute_i will stay at 0 wich will stop the while loop and start the next level
        while computes[compute_i]:
            node, compute_i = nodes_array[computes[compute_i]], compute_i + 1\

            #compute loss sums and extra sum
            len_children = children[node['children']]
            if len_children:#if no children and in compute then layer 1 loss is already set
                if len_children > 1:
                    # we need a new sidx array
                    sidx_created = False
                    node_sidx_start = sidx_ptr_i
                    node_sidx_end = 0
                    sidx_indexes[node['node_id']], sidx_i = sidx_i, sidx_i + 1
                    for p in range(node['profile_len']):
                        loss_indptr[node['loss'] + p] = loss_ptr_i
                        extras_indptr[node['extra'] + p] = extras_ptr_i

                        for c in range(node['children'] + 1, node['children'] + len_children + 1):
                            child = nodes_array[children[c]]
                            child_sidx_index = sidx_indexes[child['node_id']]
                            child_sidx_start = sidx_indptr[child_sidx_index]
                            child_sidx_end =  sidx_indptr[child_sidx_index + 1]
                            child_loss_start = loss_indptr[child['il'] + p]
                            child_extra_start = extras_indptr[child['extra'] + p]

                            for indptr in range(child_sidx_end - child_sidx_start):
                                temp_node_sidx[sidx_val[child_sidx_start + indptr]] = True
                                temp_node_loss[sidx_val[child_sidx_start + indptr]] += loss_val[child_loss_start + indptr]
                                temp_node_extras[sidx_val[child_sidx_start + indptr]] += extras_val[child_extra_start + indptr]
                        if sidx_created:
                            for node_sidx_cur in range(node_sidx_start, node_sidx_end):
                                loss_val[loss_ptr_i] = temp_node_loss[sidx_val[node_sidx_cur]]
                                temp_node_loss[sidx_val[node_sidx_cur]] = 0
                                loss_ptr_i += 1
                                extras_val[extras_ptr_i] = temp_node_extras[sidx_val[node_sidx_cur]]
                                temp_node_extras[sidx_val[node_sidx_cur]] = 0
                                extras_ptr_i += 1

                        else:
                            for sidx in [-3, -1]:
                                if temp_node_sidx[sidx]:
                                    sidx_ptr_i, loss_ptr_i, extras_ptr_i = single_array_to_sparse(sidx, sidx_ptr_i, sidx_val, temp_node_sidx,
                                                                                                   loss_ptr_i, loss_val, temp_node_loss,
                                                                                                   extras_ptr_i, extras_val, temp_node_extras)


                            for sidx in range(1, max_sidx_val+1):
                                if temp_node_sidx[sidx]:
                                    # print('res', node['node_id'], sidx, temp_node_loss[sidx], temp_node_extras[sidx])
                                    sidx_ptr_i, loss_ptr_i, extras_ptr_i = single_array_to_sparse(sidx, sidx_ptr_i, sidx_val, temp_node_sidx,
                                                                                                   loss_ptr_i, loss_val, temp_node_loss,
                                                                                                   extras_ptr_i, extras_val, temp_node_extras)
                            node_sidx_end = sidx_ptr_i
                            node_val_len = node_sidx_end - node_sidx_start
                            sidx_indptr[sidx_i] = sidx_ptr_i
                            sidx_created = True

                    #fill up all layer if necessary
                    for layer in range(node['profile_len'], node['layer_len']):
                        loss_indptr[node['loss'] + layer] = loss_indptr[node['loss']]
                        extras_indptr[node['extra'] + layer] = extras_indptr[node['extra']]

                else:
                    child = nodes_array[children[node['children'] + 1]]
                    loss_indptr[node['loss']:node['loss'] + node['layer_len']] = loss_indptr[child['il']:child['il'] + node['layer_len']]
                    sidx_indexes[node['node_id']] = sidx_indexes[child['node_id']]
                    node_sidx_start = sidx_indptr[sidx_indexes[node['node_id']]]
                    node_sidx_end = sidx_indptr[sidx_indexes[node['node_id']] + 1]
                    node_val_len = node_sidx_end - node_sidx_start

                    for p in range(node['profile_len']):
                        extras_indptr[node['extra'] + p] = extras_ptr_i
                        child_indptr = extras_indptr[child['extra'] + p]
                        extras_val[extras_ptr_i: extras_ptr_i + node_val_len] = extras_val[child_indptr: child_indptr + node_val_len]
                        extras_ptr_i += node_val_len

                    #fill up all layer if necessary
                    for layer in range(node['profile_len'], node['layer_len']):
                        extras_indptr[node['extra'] + layer] = extras_indptr[node['extra']]

            else:
                node_sidx_start = sidx_indptr[sidx_indexes[node['node_id']]]
                node_sidx_end = sidx_indptr[sidx_indexes[node['node_id']] + 1]
                node_val_len = node_sidx_end - node_sidx_start

                for p in range(node['profile_len']):
                    extras_indptr[node['extra'] + p] = extras_ptr_i
                    extras_val[extras_ptr_i: extras_ptr_i + node_val_len].fill(0)
                    extras_ptr_i += node_val_len

                #fill up all layer if necessary
                for layer in range(node['profile_len'], node['layer_len']):
                    extras_indptr[node['extra'] + layer] = extras_indptr[node['extra']]

            #compute il
            # print('length', node['node_id'], sidx_indexes[node['node_id']], sidx_i, node_val_len, sidx_indptr[sidx_indexes[node['node_id']]+1], sidx_indptr[sidx_indexes[node['node_id']]])
            for p in range(node['profile_len']):
                node_profile = node_profiles_array[node['profiles']+p]
                if node_profile['i_start'] < node_profile['i_end']:
                    loss_indptr_start =  loss_indptr[node['loss'] + p]
                    extra_indptr_start =  extras_indptr[node['extra'] + p]

                    loss_in = loss_val[loss_indptr_start: loss_indptr_start + node_val_len]
                    extra = extras_val[extra_indptr_start: extra_indptr_start + node_val_len]
                    loss_indptr[node['il'] + p] = loss_ptr_i
                    loss_ptr_i += node_val_len
                    loss_out = loss_val[loss_indptr[node['il'] + p]: loss_ptr_i]

                    for profile_index in range(node_profile['i_start'], node_profile['i_end']):
                        calc(fm_profile[profile_index],
                             loss_out,
                             loss_in,
                             extra[:, 0],
                             extra[:, 1],
                             extra[:, 2],
                             stepped)
                        # print(node['node_id'], loss_indptr[node['loss'] + p], loss_in, '=>', loss_out)
                else:
                    loss_indptr[node['il'] + p] = loss_indptr[node['loss'] + p]

            for layer in range(node['profile_len'], node['layer_len']):
                loss_indptr[node['il'] + layer] = loss_indptr[node['il']]

            if level == compute_info['max_level']:
                loss_indptr[node['ba']:node['ba'] + node['layer_len']] = loss_indptr[node['il']:node['il'] + node['layer_len']]
                if node['parent_len']:# for allocation 1 we set only the parent in computes
                    parent_id = node_parents_array[node['parent']]
                    computes[next_compute_i], next_compute_i = parent_id, next_compute_i + 1
                else:
                    computes[next_compute_i], next_compute_i = computes[compute_i - 1], next_compute_i + 1
            else:
                # set parent children and next computation
                for pa in range(node['parent_len']):
                    parent_id = node_parents_array[node['parent']+pa]
                    parent = nodes_array[parent_id]
                    parent_children_len = children[parent['children']] + 1
                    if parent_children_len == 1 and not pa:#only parent 0 is a real parent, parent 2 is only used on alloc 1
                        computes[next_compute_i], next_compute_i = parent_id, next_compute_i + 1
                    children[parent['children']] = parent_children_len
                    children[parent['children'] + parent_children_len] = computes[compute_i - 1]
        compute_i += 1

    if compute_info['allocation_rule'] == 0:
        pass
    elif compute_info['allocation_rule'] == 1:
        child_pos = np.zeros(compute_info['items_len'] + 1, dtype = np_oasis_int)
        next_compute_i += 1
        while computes[compute_i]:
            node, compute_i = nodes_array[computes[compute_i]], compute_i + 1
            len_children = children[node['children']]
            node_sidx_start = sidx_indptr[sidx_indexes[node['node_id']]]
            node_sidx_end = sidx_indptr[sidx_indexes[node['node_id']] + 1]
            node_val_len = node_sidx_end - node_sidx_start
            if len_children > 1:
                # we are summing the input loss of level 0 or 1 so there is no layer to take into account
                loss_indptr[node['loss']] = loss_ptr_i
                for node_sidx_cur in range(node_sidx_start, node_sidx_end):
                    sidx = sidx_val[node_sidx_cur]
                    loss_val[loss_ptr_i] = 0
                    for c in range(node['children'] + 1, node['children'] + len_children + 1):
                        child = nodes_array[children[c]]
                        if sidx == sidx_val[sidx_indptr[sidx_indexes[child['node_id']]] + child_pos[children[c]]]:
                            loss_val[loss_ptr_i] += loss_val[loss_indptr[child['loss']] + child_pos[children[c]]]
                            child_pos[children[c]] += 1
                    loss_ptr_i += 1

                for layer in range(1, node['layer_len']):
                    loss_indptr[node['loss'] + layer] = loss_indptr[node['loss']]

            loss_ptr_i, next_compute_i = back_allocate(node, children, nodes_array, sidx_indexes, sidx_indptr, sidx_val,
                                                       loss_indptr, loss_val, loss_ptr_i, computes, next_compute_i, 'loss', temp_node_loss)
        compute_i += 1
    else:
        for level in range(compute_info['max_level'] - compute_info['start_level']):# perform back allocation 2
            # print(level, next_compute_i, compute_i, next_compute_i-compute_i, computes[compute_i:compute_i + 2], computes[next_compute_i - 1: next_compute_i + 1])
            next_compute_i += 1
            while computes[compute_i]:
                node, compute_i = nodes_array[computes[compute_i]], compute_i + 1
                loss_ptr_i, next_compute_i = back_allocate(node,
                                                           children,
                                                           nodes_array,
                                                           sidx_indexes, sidx_indptr, sidx_val, loss_indptr, loss_val, loss_ptr_i,
                                                           computes,
                                                           next_compute_i,
                                                           'il',
                                                           temp_node_loss)
            compute_i += 1
    if net_loss:
        # we go through the last level node and replace the gross loss by net loss, then reset compute_i to its value
        out_compute_i = compute_i
        while computes[compute_i]:
            node_i, compute_i= computes[compute_i], compute_i + 1
            node = nodes_array[node_i]
            # net loss layer i is initial loss - sum of all layer up to i
            node_val_len = sidx_indptr[sidx_indexes[node['node_id']] + 1] - sidx_indptr[sidx_indexes[node['node_id']]]
            node_ba_val_prev = loss_val[loss_indptr[node['loss']]: loss_indptr[node['loss']] + node_val_len]
            for layer in range(node['layer_len']):
                node_ba_val_cur = loss_val[loss_indptr[node['ba'] + layer]: loss_indptr[node['ba'] + layer] + node_val_len]
                node_ba_val_cur[:] = np.maximum(node_ba_val_prev - node_ba_val_cur, 0)
                node_ba_val_prev = node_ba_val_cur

        compute_i = out_compute_i

    # print(compute_info['max_level'], next_compute_i, compute_i, next_compute_i-compute_i, computes[compute_i:compute_i + 2], computes[next_compute_i - 1: next_compute_i + 1])
    return compute_i


def init_variable(compute_info, max_sidx_val, temp_dir, low_memory):
    """
    extras, loss contains the same index as sidx
    therefore we can use only sidx_indexes to tract the length of each node values
    """
    max_sidx_count = max_sidx_val + EXTRA_VALUES
    len_array = max_sidx_val + 4

    if low_memory:
        sidx_val = np.memmap(os.path.join(temp_dir, "sidx_val.bin"), mode='w+',
                              shape=(compute_info['node_len'] * max_sidx_count), dtype=np_oasis_int)
        loss_val = np.memmap(os.path.join(temp_dir, "loss_val.bin"), mode='w+',
                             shape=(compute_info['loss_len'] * max_sidx_count), dtype=np_oasis_float)
        extras_val = np.memmap(os.path.join(temp_dir, "extras_val.bin"), mode='w+',
                               shape=(compute_info['extra_len'] * max_sidx_count, 3), dtype=np_oasis_float)
    else:
        sidx_val = np.zeros((compute_info['node_len'] * max_sidx_count), dtype=np_oasis_int)
        loss_val = np.zeros((compute_info['loss_len'] * max_sidx_count), dtype=np_oasis_float)
        extras_val = np.zeros((compute_info['extra_len'] * max_sidx_count, 3), dtype=np_oasis_float)

    sidx_indptr = np.zeros(compute_info['node_len'] + 1, dtype=np.int64)
    loss_indptr = np.zeros(compute_info['loss_len'] + 1, dtype=np.int64)
    extras_indptr = np.zeros(compute_info['extra_len'] + 1, dtype=np.int64)

    sidx_indexes = np.empty(compute_info['node_len'], dtype=np_oasis_int)
    children = np.zeros(compute_info['children_len'], dtype=np.uint32)
    computes = np.zeros(compute_info['compute_len'], dtype=np.uint32)

    return (max_sidx_val, max_sidx_count, len_array, sidx_indexes, sidx_indptr, sidx_val, loss_indptr, loss_val, extras_indptr, extras_val,
            children, computes)


@njit(cache=True)
def reset_variable(children, compute_i, computes):
    computes[:compute_i].fill(0)
    children.fill(0)



__all__ = [
    'get_gul_input_items',
    'write_coverages_file',
    'write_gul_input_files',
    'write_items_file',
    'write_complex_items_file'
]

import copy
import os
import sys
import warnings
import logging

from collections import OrderedDict

import numpy as np
import pandas as pd

from ..utils.coverages import SUPPORTED_COVERAGE_TYPES
from ..utils.data import (
    factorize_array,
    factorize_ndarray,
    get_dataframe,
    merge_dataframes,
    set_dataframe_column_dtypes,
)
from ..utils.defaults import (
    get_default_exposure_profile,
    GROUP_ID_COLS,
    CORRELATION_GROUP_ID,
    OASIS_FILES_PREFIXES,
)
from ..utils.exceptions import OasisException
from ..utils.fm import SUPPORTED_FM_LEVELS
from ..utils.log import oasis_log
from ..utils.defaults import (
    SOURCE_IDX,
)
from ..utils.path import as_path
from ..utils.profiles import (
    get_fm_terms_oed_columns,
    get_grouped_fm_profile_by_level_and_term_group,
    get_grouped_fm_terms_by_level_and_term_group,
    get_oed_hierarchy,
)

pd.options.mode.chained_assignment = None
warnings.simplefilter(action='ignore', category=FutureWarning)


@oasis_log
def get_gul_input_items(
    exposure_df,
    keys_df,
    exposure_profile=get_default_exposure_profile(),
    group_id_cols=['loc_id']
):
    """
    Generates and returns a Pandas dataframe of GUL input items.

    :param exposure_df: Exposure dataframe
    :type exposure_df: pandas.DataFrame

    :param keys_df: Keys dataframe
    :type keys_df: pandas.DataFrame

    :param exposure_profile: Exposure profile
    :type exposure_profile: dict

    :return: GUL inputs dataframe
    :rtype: pandas.DataFrame

    :return: Exposure dataframe
    :rtype: pandas.DataFrame
    """
    # Get the grouped exposure profile - this describes the financial terms to
    # to be found in the source exposure file, which are for the following
    # FM levels: site coverage (# 1), site pd (# 2), site all (# 3). It also
    # describes the OED hierarchy terms present in the exposure file, namely
    # portfolio num., acc. num., loc. num., and cond. num.
    profile = get_grouped_fm_profile_by_level_and_term_group(exposure_profile=exposure_profile)

    if not profile:
        raise OasisException(
            'Source exposure profile is possibly missing FM term information: '
            'FM term definitions for TIV, limit, deductible, attachment and/or share.'
        )

    # Get the OED hierarchy terms profile - this defines the column names for loc.
    # ID, acc. ID, policy no. and portfolio no., as used in the source exposure
    # and accounts files. This is to ensure that the method never makes hard
    # coded references to the corresponding columns in the source files, as
    # that would mean that changes to these column names in the source files
    # may break the method
    oed_hierarchy = get_oed_hierarchy(exposure_profile=exposure_profile)
    loc_num = oed_hierarchy['locnum']['ProfileElementName'].lower()
    acc_num = oed_hierarchy['accnum']['ProfileElementName'].lower()
    portfolio_num = oed_hierarchy['portnum']['ProfileElementName'].lower()

    # The (site) coverage FM level ID (# 1 in the OED FM levels hierarchy)
    cov_level_id = SUPPORTED_FM_LEVELS['site coverage']['id']

    # Get the TIV column names and corresponding coverage types
    tiv_terms = OrderedDict({v['tiv']['CoverageTypeID']: v['tiv']['ProfileElementName'].lower() for k, v in profile[cov_level_id].items()})
    tiv_cols = list(tiv_terms.values())

    # Get the list of coverage type IDs - financial terms for the coverage
    # level are grouped by coverage type ID in the grouped version of the
    # exposure profile (profile of the financial terms sourced from the
    # source exposure file)
    cov_types = [v['id'] for v in SUPPORTED_COVERAGE_TYPES.values()]

    # Get the FM terms profile (this is a simplfied view of the main grouped
    # profile, containing only information about the financial terms), and
    # the list of OED colum names for the financial terms for the site coverage
    # (# 1 ) FM level
    fm_terms = get_grouped_fm_terms_by_level_and_term_group(grouped_profile_by_level_and_term_group=profile)
    terms_floats = ['deductible', 'deductible_min', 'deductible_max', 'limit']
    terms_ints = ['ded_code', 'ded_type', 'lim_code', 'lim_type']
    terms = terms_floats + terms_ints
    term_cols_floats = get_fm_terms_oed_columns(
        fm_terms,
        levels=['site coverage'],
        term_group_ids=cov_types,
        terms=terms_floats
    )
    term_cols_ints = get_fm_terms_oed_columns(
        fm_terms,
        levels=['site coverage'],
        term_group_ids=cov_types,
        terms=terms_ints
    )
    term_cols = term_cols_floats + term_cols_ints

    # Create the basic GUL inputs dataframe from merging the exposure and
    # keys dataframes on loc. number/loc. ID; filter out any rows with
    # zeros for TIVs for all coverage types, and replace any nulls in the
    # cond.num. and TIV columns with zeros

    # Select only the columns required. This reduces memory use significantly for portfolios
    # that include many OED columns.
    exposure_df_gul_inputs_cols = ['loc_id', portfolio_num, acc_num, loc_num] + term_cols + tiv_cols
    if SOURCE_IDX['loc'] in exposure_df:
        exposure_df_gul_inputs_cols += [SOURCE_IDX['loc']]
    # Remove any duplicate column names used to assign group_id
    group_id_cols = list(set(group_id_cols))

    # Ignore any column names used to assign group_id that are missing or not supported
    # Valid group id columns can be either
    # 1. exist in the location file
    # 2. be listed as a useful internal col
    valid_oasis_group_cols = [
        'item_id',
        'peril_id',
        'coverage_id',
        'coverage_type_id',
    ]
    for col in group_id_cols:
        if col not in list(exposure_df.columns) + valid_oasis_group_cols:
            warnings.warn('Column {} not found in loc file, or a valid internal oasis column'.format(col))
            group_id_cols.remove(col)

    # Should list of column names used to group_id be empty, revert to
    # default
    if len(group_id_cols) == 0:
        group_id_cols = GROUP_ID_COLS

    # Only add group col if not internal oasis col
    missing_group_id_cols = []
    for col in group_id_cols:
        if col in valid_oasis_group_cols:
            pass
        elif col not in exposure_df_gul_inputs_cols:
            missing_group_id_cols.append(col)
    exposure_df_gul_inputs_cols += missing_group_id_cols

    # Check if correlation group field used to drive group id
    # and test that it's present and poulated with integers
    correlation_group_id = list(map(lambda col: col.lower(), CORRELATION_GROUP_ID))
    correlation_field = correlation_group_id[0]
    correlation_check = False
    if group_id_cols == correlation_group_id:
        if correlation_field in exposure_df.columns:
            if exposure_df[correlation_field].astype('uint32').isnull().sum() == 0:
                correlation_check = True

    query_nonzero_tiv = " | ".join(f"({tiv_col} != 0)" for tiv_col in tiv_cols)
    exposure_df.loc[:, tiv_cols] = exposure_df.loc[:, tiv_cols].fillna(0.0)
    exposure_df.query(query_nonzero_tiv, inplace=True, engine='numexpr')


    gul_inputs_df = exposure_df[exposure_df_gul_inputs_cols]
    gul_inputs_df.drop_duplicates('loc_id', inplace=True, ignore_index=True)

    # Rename the main keys dataframe columns - this is due to the fact that the
    # keys file headers use camel case, and don't use underscored names, which
    # is the convention used for the GUL and IL inputs dataframes in the MDK
    keys_df.rename(
        columns={
            'locid': 'loc_id' if 'loc_id' not in keys_df else 'locid',
            'perilid': 'peril_id',
            'coveragetypeid': 'coverage_type_id',
            'areaperilid': 'areaperil_id',
            'vulnerabilityid': 'vulnerability_id',
            'modeldata': 'model_data'
        },
        inplace=True,
        copy=False  # Pandas copies column data by default on rename
    )

    # If the keys file relates to a complex/custom model then look for a
    # ``modeldata`` column in the keys file, and ignore the area peril
    # and vulnerability ID columns
    if 'model_data' in keys_df:
        keys_df['areaperil_id'] = keys_df['vulnerability_id'] = -1

    gul_inputs_df = merge_dataframes(
        gul_inputs_df,
        keys_df,
        join_on='loc_id',
        how='inner'
    )

    if gul_inputs_df.empty:
        raise OasisException(
            'Inner merge of the exposure file dataframe '
            'and the keys file dataframe on loc. number/loc. ID '
            'is empty - '
            'please check that the loc. number and loc. ID columns '
            'in the exposure and keys files respectively have a non-empty '
            'intersection'
        )

    # Free memory after merge, before memory-intensive restructuring of data
    del keys_df

    gul_inputs_df[tiv_terms.values()].fillna(0, inplace=True)

    # Group the rows in the GUL inputs table by coverage type, and set the
    # IL terms (and BI coverage boolean) in each group and update the
    # corresponding frame section in the GUL inputs table
    gul_inputs_reformatted_chunks = []
    for cov_type, cov_type_group in gul_inputs_df.groupby(by=['coverage_type_id'], sort=True):
        tiv_col = tiv_terms[cov_type]

        # Remove rows with null/0 TIV
        cov_type_group.query(f"{tiv_col} > 0.0", inplace=True)

        # Drop columns corresponding to other cov types
        other_cov_types = [v['id'] for v in SUPPORTED_COVERAGE_TYPES.values() if v['id'] != cov_type]
        other_cov_type_term_cols = (
            [v for k, v in tiv_terms.items() if k != cov_type] +
            get_fm_terms_oed_columns(fm_terms=fm_terms, levels=['site coverage'], term_group_ids=other_cov_types, terms=terms)
        )
        other_tiv_cols = list(set(tiv_terms.values()) - {tiv_col})
        cov_type_group.drop(
            columns=other_cov_type_term_cols + other_tiv_cols,
            errors="ignore",  # Ignore if any of these cols don't exist
            inplace=True
        )

        cov_type_group['is_bi_coverage'] = cov_type == SUPPORTED_COVERAGE_TYPES['bi']['id']

        # Convert null T&C values to 0
        # Move all T&C values from coverage-specific columns to generic columns
        # One row with four coverages is being transformed to four rows, one per coverage,
        # in this loop
        cov_type_terms = [t for t in terms if fm_terms[cov_level_id][cov_type].get(t)]
        cov_type_term_cols = get_fm_terms_oed_columns(fm_terms, levels=['site coverage'], term_group_ids=[cov_type], terms=cov_type_terms)
        column_mapping_dict = {
            generic_col: cov_col
            for generic_col, cov_col in zip(cov_type_term_cols, cov_type_terms)
        }
        column_mapping_dict[tiv_col] = 'tiv'
        cov_type_group.rename(columns=column_mapping_dict, inplace=True, copy=False)
        cov_type_group.loc[:, ['tiv'] + cov_type_terms].fillna(0.0, inplace=True)

        cov_type_group['coverage_type_id'] = cov_type

        gul_inputs_reformatted_chunks.append(cov_type_group)

    # Concatenate chunks. Sort by index to preserve item_id order in generated outputs compared
    # to original code.
    gul_inputs_df = pd.concat(gul_inputs_reformatted_chunks).sort_index().reset_index()
    # Set default values and data types for BI coverage boolean, TIV, deductibles and limit
    dtypes = {
        **{t: 'uint8' for t in term_cols_ints + terms_ints},
        **{'is_bi_coverage': 'bool'}
    }
    gul_inputs_df = set_dataframe_column_dtypes(gul_inputs_df, dtypes)

    if gul_inputs_df.empty:
        raise OasisException(
            'Empry gul_inputs_df dataframe after dropping rows with zero for tiv, '
            'please check the exposure input files'
        )

    # Set the item IDs and coverage IDs, and defaults and data types for
    # layer IDs and agg. IDs
    item_ids = gul_inputs_df.index + 1
    gul_inputs_df['coverage_id'] = factorize_ndarray(
        gul_inputs_df.loc[:, ['loc_id', 'coverage_type_id']].values, col_idxs=range(2))[0]
    gul_inputs_df = gul_inputs_df.assign(
        item_id=item_ids,
    )
    dtypes = {
        **{t: 'uint32' for t in ['item_id', 'coverage_id']},
        **{t: 'uint8' for t in terms_ints}
    }
    gul_inputs_df = set_dataframe_column_dtypes(gul_inputs_df, dtypes)


    # Set the group ID
    # If the group id is set according to the correlation group field then map this field
    # directly, otherwise create an index of the group id fields
    group_id_cols.sort()
    if correlation_check == True:
        gul_inputs_df['group_id'] = gul_inputs_df[correlation_group_id]
    else:
        if len(group_id_cols) > 1:
            gul_inputs_df['group_id'] = factorize_ndarray(
                gul_inputs_df.loc[:, group_id_cols].values,
                col_idxs=range(len(group_id_cols)),
                sort_opt=True
            )[0]
        else:
            gul_inputs_df['group_id'] = factorize_array(
                gul_inputs_df[group_id_cols[0]].values
            )[0]
    gul_inputs_df['group_id'] = gul_inputs_df['group_id'].astype('uint32')



    # Select only required columns
    # Order here matches test output expectations
    usecols = (
        ['loc_id', portfolio_num, acc_num, loc_num] +
        ([SOURCE_IDX['loc']] if SOURCE_IDX['loc'] in gul_inputs_df else []) +
        ['peril_id', 'coverage_type_id', 'tiv', 'areaperil_id', 'vulnerability_id'] +
        terms +
        (['model_data'] if 'model_data' in gul_inputs_df else []) +
        ['is_bi_coverage', 'group_id', 'coverage_id', 'item_id', 'status']
    )
    usecols = [col for col in usecols if col in gul_inputs_df]
    gul_inputs_df = gul_inputs_df[usecols]

    return gul_inputs_df


@oasis_log
def write_complex_items_file(gul_inputs_df, complex_items_fp, chunksize=100000):
    """
    Writes a complex model items file.

    :param gul_inputs_df: GUL inputs dataframe
    :type gul_inputs_df: pandas.DataFrame

    :param complex_items_fp: Complex/custom model items file path
    :type complex_items_fp: str

    :return: Complex/custom model items file path
    :rtype: str
    """
    try:
        gul_inputs_df.loc[:, ['item_id', 'coverage_id', 'model_data', 'group_id']].drop_duplicates().to_csv(
            path_or_buf=complex_items_fp,
            encoding='utf-8',
            mode=('w' if os.path.exists(complex_items_fp) else 'a'),
            chunksize=chunksize,
            index=False
        )
    except (IOError, OSError) as e:
        raise OasisException("Exception raised in 'write_complex_items_file'", e)


@oasis_log
def write_items_file(gul_inputs_df, items_fp, chunksize=100000):
    """
    Writes an items file.

    :param gul_inputs_df: GUL inputs dataframe
    :type gul_inputs_df: pandas.DataFrame

    :param items_fp: Items file path
    :type items_fp: str

    :return: Items file path
    :rtype: str
    """
    try:
        gul_inputs_df.loc[:, ['item_id', 'coverage_id', 'areaperil_id', 'vulnerability_id', 'group_id']].drop_duplicates().to_csv(
            path_or_buf=items_fp,
            encoding='utf-8',
            mode=('w' if os.path.exists(items_fp) else 'a'),
            chunksize=chunksize,
            index=False
        )
    except (IOError, OSError) as e:
        raise OasisException("Exception raised in 'write_items_file'", e)

    return items_fp


@oasis_log
def write_coverages_file(gul_inputs_df, coverages_fp, chunksize=100000):
    """
    Writes a coverages file.

    :param gul_inputs_df: GUL inputs dataframe
    :type gul_inputs_df: pandas.DataFrame

    :param coverages_fp: Coverages file path
    :type coverages_fp: str

    :return: Coverages file path
    :rtype: str
    """
    try:
        gul_inputs_df.loc[:, ['coverage_id', 'tiv']].drop_duplicates().to_csv(
            path_or_buf=coverages_fp,
            encoding='utf-8',
            mode=('w' if os.path.exists(coverages_fp) else 'a'),
            chunksize=chunksize,
            index=False
        )
    except (IOError, OSError) as e:
        raise OasisException("Exception raised in 'write_coverages_file'", e)

    return coverages_fp


@oasis_log
def write_gul_input_files(
    gul_inputs_df,
    target_dir,
    oasis_files_prefixes=copy.deepcopy(OASIS_FILES_PREFIXES['gul']),
    chunksize=(2 * 10 ** 5)
):
    """
    Writes the standard Oasis GUL input files to a target directory, using a
    pre-generated dataframe of GUL input items. The files written are
    ::

        items.csv
        coverages.csv

    and optionally a complex items file in case of a complex/custom model.

    :param gul_inputs_df: GUL inputs dataframe
    :type gul_inputs_df: pandas.DataFrame

    :param target_dir: Target directory in which to write the files
    :type target_dir: str

    :param oasis_files_prefixes: Oasis GUL input file name prefixes
    :param oasis_files_prefixes: dict

    :param chunksize: The chunk size to use when writing out the
                      input files
    :type chunksize: int

    :return: GUL input files dict
    :rtype: dict
    """
    # Clean the target directory path
    target_dir = as_path(target_dir, 'Target IL input files directory', is_dir=True, preexists=False)

    # Set chunk size for writing the CSV files - default is the minimum of 100K
    # or the GUL inputs frame size
    chunksize = chunksize or min(chunksize, len(gul_inputs_df))

    # If no complex model data present then remove the corresponding file
    # name from the files prefixes dict, which is used for writing the
    # GUl input files
    if 'model_data' not in gul_inputs_df:
        if oasis_files_prefixes.get('complex_items'):
            oasis_files_prefixes.pop('complex_items')

    # A dict of GUL input file names and file paths
    gul_input_files = {
        fn: os.path.join(target_dir, '{}.csv'.format(oasis_files_prefixes[fn]))
        for fn in oasis_files_prefixes
    }
    this_module = sys.modules[__name__]
    # Write the files serially
    for fn in gul_input_files:
        getattr(this_module, 'write_{}_file'.format(fn))(gul_inputs_df.copy(deep=True), gul_input_files[fn], chunksize)

    return gul_input_files
